package de.ordix.onews.persistence.impl;

import java.io.Serializable;
import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Root;

import de.ordix.onews.persistence.dao.Dao;
import de.ordix.onews.persistence.util.NoResultDaoException;
import de.ordix.onews.persistence.util.PredicateBuilder;

/**
* Licensed to the Apache Software Foundation (ASF) under one or more
* contributor license agreements. See the NOTICE file distributed with
* this work for additional information regarding copyright ownership.
* The ASF licenses this file to You under the Apache License, Version 2.0
* (the "License"); you may not use this file except in compliance with
* the License. You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
* 
 * Enthält die Implementierungen der allgemeingueltigen Dao "Dao"
 * 
 * @author sk@ordix.de
 *
 */
public abstract class GenericDaoImpl<E, P extends Serializable> implements Dao<E, P> {

    @PersistenceContext(unitName = "zmapds")
    protected EntityManager entityManager;
    private final Class<E> type;

    public GenericDaoImpl(Class<E> clazz) {
        this.type = clazz;
    }

    protected void detach() {
        entityManager.flush();
        entityManager.clear();
    }

    @Override
    public E save(E object) {
        entityManager.persist(object);
        detach();
        return object;

    }
    @Override
    public void delete(E object) {
        E foundObject = entityManager.merge(object);
        entityManager.remove(foundObject);
        detach();
    }

    @Override
    public E getById(P id) throws NoResultDaoException {
        E result = entityManager.find(type, id);
        detach();
        if (result == null) {
            throw new NoResultDaoException();
        }
        return result;
    }

    @Override
    public E update(E object) {
        E returnValue = entityManager.merge(object);
        detach();
        return returnValue;

    }

    @Override
    public List<E> selectAll() throws NoResultDaoException {
        CriteriaBuilder criteriaBuilder = entityManager.getCriteriaBuilder();
        CriteriaQuery<E> criteriaQuery = criteriaBuilder.createQuery(type);
        Root<E> rootEntry = criteriaQuery.from(type);
        CriteriaQuery<E> all = criteriaQuery.select(rootEntry);
        TypedQuery<E> allQuery = entityManager.createQuery(all);
        detach();
        List<E> resultList = allQuery.getResultList();
        if (resultList == null || resultList.isEmpty()) {
            throw new NoResultDaoException();
        }
        return resultList;
    }

  

    @Override
    public void setEntityManager(EntityManager entityManager) {
        this.entityManager = entityManager;
    }
    
    /**
     * @throws NoResultDaoException
     * @see de.postbank.zmap.persistenz.dao.Dao
     */
    @Override
    public List<E> findByExample(Object example) throws NoResultDaoException {
        TypedQuery<E> typedQuery = getTypedQuery4findByExmaple(example);
        List<E> resultList = typedQuery.getResultList();
        detach();
        if (resultList == null || resultList.isEmpty()) {
            throw new NoResultDaoException();
        }
        return resultList;
    }
    
   
    private TypedQuery<E> getTypedQuery4findByExmaple(Object example) {
        CriteriaBuilder criteriaBuilder = entityManager.getCriteriaBuilder();
        CriteriaQuery<E> criteriaQuery = new PredicateBuilder<>(type).findByObject(example, criteriaBuilder);
        return entityManager.createQuery(criteriaQuery);
    }

}