package de.ordix.onews.persistence.util;

import java.text.SimpleDateFormat;
import java.time.DateTimeException;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Locale;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeConstants;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;

/**
* Licensed to the Apache Software Foundation (ASF) under one or more
* contributor license agreements. See the NOTICE file distributed with
* this work for additional information regarding copyright ownership.
* The ASF licenses this file to You under the Apache License, Version 2.0
* (the "License"); you may not use this file except in compliance with
* the License. You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
* 
 * Beinhaltet Hilfsfunktionen die eine Datum-Eingabe bearbeiten. Es werden
 * jegliche Umwandlungen, Formatierungen und ähnliches mit einem Datum
 * unterstützt.
 * 
 * @author Dawid Thomanek
 *
 */
public class DatumFormatter {
    static final String ZEIT_ZONE = "Z"; // "Z" = UTC

    public enum DatumsFormat {
        DEUTSCH_TAGESDATUM("dd.MM.yyyy"), TECHNISCH("yyyyMMdd");
        DatumsFormat(String format) {
            formatString = format;
        }

        String formatString;

        public String getFormatString() {
            return formatString;
        }
    }

    static final String DATE_FORMAT_TAGESDATUM = "dd.MM.yyyy";

    /**
     * Konvertiert einen String der Form <code>DD.MM.YYYY</code> in ein
     * Date-Objekt um. Es wird die Zeitzone <code>UTF</code> benutzt.
     * 
     * @param datum
     *            Datum als eine Zeichenkette im Format "DD.MM.YYYY"
     * @return Objekt mit den Datumsangaben aus <code>datum</code>
     */
    public static Date toDate(String datum) {
        return toDate(datum, ZEIT_ZONE);
    }

    /**
     * Konvertiert einen String der Form <code>DD.MM.YYYY</code> in ein
     * Date-Objekt um.
     * 
     * @param datum
     *            Datum als eine Zeichenkette im Format "DD.MM.YYYY"
     * @param zeitZoneOffset
     *            Zeitzone in der das Datumsformat ausgegeben wird
     * @return Objekt mit den Datumsangaben aus <code>datum</code>
     */
    public static Date toDate(String datum, final String zeitZoneOffset) {
        final String datumsformatMuster = "DD.MM.YYYY";

        // ein leerer String gibt ein leeres Datum zurück
        if (datum == null || datum.isEmpty()) {
            return null;
        }

        if (!datum.matches("[0-9\\.]{" + datumsformatMuster.length() + "}")) {
            throw new DateTimeException("Das Datumformat ist falsch");
        }

        LocalDateTime lDate = LocalDateTime.of(Integer.valueOf(datum.substring(6, 10)),
                Integer.valueOf(datum.substring(3, 5)), Integer.valueOf(datum.substring(0, 2)), 0, 0);

        Date datumObject = new Date();
        datumObject.setTime(lDate.toEpochSecond(ZoneOffset.of(zeitZoneOffset)) * 1000);

        return datumObject;
    }

    /**
     * Prüft, ob ein String folgendem Datumsformat entspricht: "yyyymmddhhmm"
     * 
     * @param substring
     *            zu prüfender String
     * @return <code>true</code> falls das format korrekt ist <code>false</code>
     *         falls das format nicht dem muster entspricht
     */
    public static boolean parseDateFormat1(String substring) {
        String datumsformatMuster = "yyyymmddhhmm";

        if (!substring.matches("[0-9]{" + datumsformatMuster.length() + "}")) {
            return false;
        }

        // prüfe das Datumsformat
        if (Integer.parseInt(substring.substring(0, 4)) > 0 // Jahr
                && 0 < Integer.parseInt(substring.substring(4, 6)) // Monat
                && Integer.parseInt(substring.substring(4, 6)) < 13 // Monat
                && 0 < Integer.parseInt(substring.substring(6, 8)) // Tag
                && Integer.parseInt(substring.substring(6, 8)) < 32 // Tag
                && Integer.parseInt(substring.substring(8, 10)) < 24 // Stunde
                && Integer.parseInt(substring.substring(10, 12)) < 60) // Minute
        {
            return true;
        }

        return false;
    }

    public static boolean parseDateYearAndMonth(String string) {
        String datumsFormatMuster = "YYMM";
        if (!string.matches("[0-9]{" + datumsFormatMuster.length() + "}")) {
            return false;
        }

        if (Integer.parseInt(string.substring(0, 2)) > 0 && Integer.parseInt(string.substring(2, 4)) > 0) {
            return true;
        }
        return false;
    }

    /**
     * Konvertiert ein Date-Objekt in ein XMLGregorianCalendar-Objekt
     * 
     * @param datum
     *            das Datum das konvertiert wird
     * @return das angegebene Datum als XMLGregorianCalendar-Objekt
     */
    public XMLGregorianCalendar konvertiereDateZuXMLGregorianCalendar(Date datum) {
        XMLGregorianCalendar xmlDate = null;

        GregorianCalendar cal = new GregorianCalendar();
        cal.setTime(datum);
        try {
            xmlDate = DatatypeFactory.newInstance().newXMLGregorianCalendarDate(cal.get(Calendar.YEAR),
                    cal.get(Calendar.MONTH) + 1, cal.get(Calendar.DAY_OF_MONTH), DatatypeConstants.FIELD_UNDEFINED);
        } catch (DatatypeConfigurationException e) {
            throw new TechnicalException("Das Datum konnte nicht konvertiert werden", e);
        }

        return xmlDate;
    }

    /**
     * Wandelt Zeichenkette in LocalDate um
     * 
     * @param datumsString
     * @return
     */
    public static LocalDate getLocalDate(String datumsString) {
        return getLocalDate(datumsString, DatumsFormat.DEUTSCH_TAGESDATUM);
    }

    public static LocalDate getLocalDate(String datumsString, DatumsFormat datumsFormat) {
        if (datumsString == null) {
            return null;
        }
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern(datumsFormat.getFormatString())
                .withLocale(Locale.GERMANY);
        LocalDate localDate = LocalDate.parse(datumsString, formatter);
        return localDate;
    }

    public static LocalDateTime getLocalDateTime(String datumsString) {
        return getLocalDateTime(datumsString, DatumsFormat.DEUTSCH_TAGESDATUM);
    }

    public static LocalDateTime getLocalDateTime(String datumsString, DatumsFormat datumsFormat) {
        if (datumsString == null) {
            return null;
        }
        LocalDate localDate = getLocalDate(datumsString, datumsFormat);
        return localDate.atStartOfDay();
    }

    public static String formatiereDatum(LocalDate datum) {
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("dd.MM.yyyy");
        return datum.format(formatter);
    }

    public static String formatiereDatum(LocalDateTime datum) {
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("dd.MM.yyyy");
        return datum.format(formatter);
    }

    public static String dateToString(Date datum) {
        SimpleDateFormat formatter = new SimpleDateFormat("dd.MM.yyyy");
        return formatter.format(datum);
    }

    /**
     * Formatiert ein Datum in einen String mit dem Format 'yyyyMMdd'
     * 
     * @param datum
     * @return
     */
    public static String dateToStringForDateiname(Date datum) {
        SimpleDateFormat formatter = new SimpleDateFormat("yyyyMMdd");
        return formatter.format(datum);
    }

}
