package de.ordix.onews.persistence.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;

import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Expression;
import javax.persistence.criteria.Path;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;


/**
* Licensed to the Apache Software Foundation (ASF) under one or more
* contributor license agreements. See the NOTICE file distributed with
* this work for additional information regarding copyright ownership.
* The ASF licenses this file to You under the Apache License, Version 2.0
* (the "License"); you may not use this file except in compliance with
* the License. You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
* 
 * Diese Klasse unterstützt die Suche nach einem Objekt nach dem Muster Find-By-Example
 * 
 * @author sk@ordix.de
 *
 */
@SuppressWarnings("rawtypes")
public class PredicateBuilder<E> {
    // Sonarqube: wont fix

    static final Logger LOGGER4J = LogManager.getLogger(PredicateBuilder.class.getName());

    private Class<E> type;
    private Map<String, Class> targetTypes = null;
    Queue<ZmapCriteria> criteriaQueue = new LinkedList<>();

    public PredicateBuilder(Class<E> clazz) {
        type = clazz;
        targetTypes = getTargetTypes(type);
    }

    private enum Operator {
        NONE, EQUAL, LIKE, LESS, GREATER, LESSOREQUAL, GREATEROREQUAL, BETWEEN, NULL;

        /**
         * Gibt je nach Anfang des Strings einen Operator zurück
         * 
         * > Operator.GREATER >= Operator.GREATEROREQUAL < Operator.LESS <=
         * Operator.LESSOREQUAL
         * 
         * @param string
         * @return
         */
        static Operator getPrefixOperator(String string) {
            if (string.startsWith("<")) {
                if (string.charAt(1) == '=') {
                    return Operator.LESSOREQUAL;
                } else {
                    return Operator.LESS;
                }
            } else if (string.startsWith(">")) {
                if (string.charAt(1) == '=') {
                    return Operator.GREATEROREQUAL;
                } else {
                    return Operator.GREATER;
                }
            }
            return Operator.NONE;
        }

        /**
         * Gibt einen enthaltenen Operator (LIKE %) zurück, wenn er enthalten
         * ist
         * 
         * @param string
         * @return
         */
        static Operator getInfixOperator(String string) {
            if (string.contains("%") && string.split("%").length >= string.split("\\%").length) {
                return Operator.LIKE;
            }
            return Operator.NONE;
        }

        /**
         * Entfernt <, <=, >=, > vom Anfang eines Strings
         * 
         * @param string
         * @return
         */
        static String removePrefixOperator(String string) {
            String trimmed = string;
            if (trimmed.startsWith(">") || trimmed.startsWith("<")) {
                trimmed = trimmed.substring(1);
            }
            if (trimmed.startsWith("=")) {
                trimmed = trimmed.substring(1);
            }

            return trimmed;
        }
    }

    /**
     * Hilfsklasse, um eine Bedingung zwischenzuspeichern
     * 
     * @author philipp
     *
     */
    @SuppressWarnings("unchecked")
    class ZmapCriteria {
        Expression attribute;
        Operator comparator;
        Object search;
        Class targetType;

        ZmapCriteria(Expression attribute, Operator comparator, Object search, Class targetType) {
            this.attribute = attribute;
            this.comparator = comparator;
            this.search = search;
            this.targetType = targetType;
            criteriaQueue.add(this);
        }

        Predicate buildPredicate(CriteriaBuilder criteriaBuilder) {
            if (Operator.NULL.equals(comparator)) {
                return buildPredicateIsNull(criteriaBuilder);
            }
            if (targetType == String.class) {
                return buildPredicateForString(criteriaBuilder);
            } else if (Comparable.class.isAssignableFrom(targetType)) {
                return buildPredicateForComparable(criteriaBuilder);
            } else {
                return criteriaBuilder.equal(attribute, search);
            }
        }

        /**
         * Erstellt ein Predicate für die Überprüfung auf null
         * @param criteriaBuilder
         * @return
         */
        private Predicate buildPredicateIsNull(CriteriaBuilder criteriaBuilder) {
            return criteriaBuilder.isNull(attribute);
        }

        /**
         * Setzt Prädikate eines Strings zusammen
         * 
         * @param criteriaBuilder
         * @return
         */
        private Predicate buildPredicateForString(CriteriaBuilder criteriaBuilder) {
            if (comparator.equals(Operator.LIKE)) {
                return criteriaBuilder.like(attribute, (String) search);
            }
            return criteriaBuilder.equal(attribute, search);
        }

        /**
         * Setzt Prädikate eines Comparables zusammen, zB Daten, Zahlen, ...
         * 
         * @param criteriaBuilder
         * @return
         */
        private Predicate buildPredicateForComparable(CriteriaBuilder criteriaBuilder) {
            if (!(search instanceof String)) { // muss ein enum sein
                return criteriaBuilder.equal(attribute, search);
            }
            if (comparator == Operator.BETWEEN) {
                Comparable[] borders = intervalFromString((String) search, targetType);
                return criteriaBuilder.between(attribute, borders[0], borders[1]);
            }

            Comparable comparableSearch = valueFromString((String) search, targetType);
            if (comparator == Operator.LESS) {
                return criteriaBuilder.lessThan(attribute, comparableSearch);
            }
            if (comparator == Operator.GREATER) {
                return criteriaBuilder.greaterThan(attribute, comparableSearch);
            }
            if (comparator == Operator.LESSOREQUAL) {
                return criteriaBuilder.lessThanOrEqualTo(attribute, comparableSearch);
            }
            if (comparator == Operator.GREATEROREQUAL) {
                return criteriaBuilder.greaterThanOrEqualTo(attribute, comparableSearch);
            }

            return criteriaBuilder.equal(attribute, comparableSearch);
        }

        private Comparable[] intervalFromString(String s, Class targetType) {
            String[] split = s.split("-");
            String leftString = split[0];
            String rightString = split[1];
            if (Number.class.isAssignableFrom(targetType)) {
                if (split.length == 3 && "".equals(split[0])) {// -x-x
                    leftString = "-" + split[1];
                    rightString = split[2];
                } else if (split.length == 3) {// x--x
                    leftString = split[0];
                    rightString = "-" + split[2];
                } else if (split.length == 4) {// -x--x
                    leftString = "-" + split[1];
                    rightString = "-" + split[3];
                } else if (split.length > 4) {// ungueltig
                    throw new IllegalStateException("Ungueltiges Zeichen: '-'");
                }
            } else {
                if (split.length > 2)
                    throw new IllegalStateException("Ungueltiges Zeichen: '-'");
            }

            return valuesFromStrings(leftString, rightString, targetType);
        }

        /**
         * Gibt ein Array mit zwei Werten zurück, von denen das 0-te Element die
         * linke Grenze und das 1-te Element die rechte Grenze eines Intervalls
         * ist
         */
        private Comparable valueFromString(String string, Class<? extends Comparable> targetType) {
            Comparable returnValue = null;

            // Konvertierung in die entsprechenden Typen. Geht das auch
            // einfacher?
            if (targetType == Date.class) {
                returnValue = convertToDate(string);
            } else if (targetType == Long.class) {
                returnValue = Long.valueOf(string);
            } else if (targetType == Integer.class) {
                returnValue = Integer.valueOf(string);
            } else if (targetType == LocalDate.class) {
                returnValue = DatumFormatter.getLocalDate(string);
            }  else if (targetType == LocalDateTime.class) {
                returnValue = DatumFormatter.getLocalDateTime(string);
            } else {
                throw new TechnicalException("Cannot convert into target type " + targetType.getName());
            }

            return returnValue;
        }

        /**
         * Gibt ein Array mit zwei Werten zurück, von denen das 0-te Element die
         * linke Grenze und das 1-te Element die rechte Grenze eines Intervalls
         * ist
         */
        private Comparable[] valuesFromStrings(String leftString, String rightString,
                Class<? extends Comparable> targetType) {
            Comparable[] array = null;

            // Konvertierung in die entsprechenden Typen. Geht das auch
            // einfacher?
            if (targetType == Date.class) {
                array = new Date[] { convertToDate(leftString), convertToDate(rightString) };
            } else if (targetType == LocalDateTime.class) {
                array = new LocalDateTime[] { convertToDateTime(leftString), convertToDateTime(rightString) }; 
            } else if (targetType == LocalDate.class) {
                array = new LocalDate[] { convertToLocalDate(leftString), convertToLocalDate(rightString) }; 
            }else if (targetType == Long.class) {
                array = new Long[] { Long.valueOf(leftString), Long.valueOf(rightString) };
            } else if (targetType == Integer.class) {
                array = new Integer[] { Integer.valueOf(leftString), Integer.valueOf(rightString) };
            } else {
                throw new TechnicalException("Type " + targetType.getName() + " cannot been converted");
            }

            if (array[1].compareTo(array[0]) < 0) {// Ausschlussintervall
                                                   // verhindern
                Comparable tmp = array[0];
                array[0] = array[1];
                array[1] = tmp;
            }

            return array;
        }

        private LocalDate convertToLocalDate(String leftString) {
            return DatumFormatter.getLocalDate(leftString);
        }
        
        private LocalDateTime convertToDateTime(String leftString) {
            return DatumFormatter.getLocalDateTime(leftString);
        }

        private Date convertToDate(String datum) {
            SimpleDateFormat dateFormat = new SimpleDateFormat("dd.MM.yyyy");
            Date returnValue = null;
            try {
                returnValue = dateFormat.parse(datum);
            } catch (ParseException e) {
                LOGGER4J.info(e.getMessage());
            }
            return returnValue;
        }
    }

    /**
     * Geliefert wird eine CriteriaQuery, die dem Example-Objekt entspricht.
     * 
     * Voraussetzung: CriteriaBuilder criteriaBuilder =
     * entityManager.getCriteriaBuilder(); Das Ergebnis (all) kann
     * beispielsweise verwendet werden: TypedQuery<E> allQuery =
     * entityManager.createQuery(all); allquery.getRusltList();
     * 
     * @param exampleObject
     *            Suchobjekt
     * @return CriteriaQuery, die
     */
    public CriteriaQuery<E> findByObject(Object exampleObject, CriteriaBuilder criteriaBuilder) {
        CriteriaQuery<E> criteriaQuery = criteriaBuilder.createQuery(type);
        Root<E> rootEntry = criteriaQuery.from(type);
        CriteriaQuery<E> all = criteriaQuery.select(rootEntry);
        Predicate predicate = findPredicate(exampleObject, criteriaBuilder, rootEntry);
        criteriaQuery.where(predicate);
        return all;
    }

    /**
     * 
     * @param exampleObject
     * @param criteriaBuilder
     * @param criteriaQuery
     *            = criteriaBuilder.createQuery(type);
     * @param rootEntry
     *            = criteriaQuery.from(type);
     * @return Predicate auf Grundlage des Example-Objekts
     */
    public Predicate findPredicate(Object exampleObject, CriteriaBuilder criteriaBuilder
           , Root<E> rootEntry) {

        Predicate predicate;
        List<Predicate> predicateList = new ArrayList<>();

        addCriteria(rootEntry, exampleObject, targetTypes);// Hinzufuegen des
                                                           // exampleObjects

        for (ZmapCriteria c : criteriaQueue) {
            predicateList.add(c.buildPredicate(criteriaBuilder));
        }

        if (predicateList.isEmpty()) {
            LOGGER4J.warn("Es wurden keine Suchbegriffe angegeben.", new Exception());
            throw new KeineSuchbegriffeAngegebenException("Es wurden keine Suchbegriffe angegeben.");
        }

        Predicate[] predicateArray = new Predicate[1];
        predicate = criteriaBuilder.and(predicateList.toArray(predicateArray));
        return predicate;
    }

    /**
     * Fügt eine Bedingung hinzu (String)
     * 
     * @param property: Property für die Bedingung
     * @param value: Angegebener Suchbegriff
     * @param targetType: Datentyp der entity für diese Property
     * @return
     */
    protected ZmapCriteria addCriteria4String(Path<E> property, String value, Class targetType) {
        // Operationen fuer Objekte, die vergleichbar sind
        if ("${null}".equals(value)) {
            return new ZmapCriteria(property, Operator.NULL, value, targetType);
        }
        if (targetType != String.class && Comparable.class.isAssignableFrom(targetType)) {
            if (value.matches("[^<>=]+-.+")) {// Wertebereich, z.B.
                // 01.01.2016-01.03.2016,
                // oder 1-6
                return new ZmapCriteria(property, Operator.BETWEEN, value, targetType);
            } else {// Vergleichsoperatoren
                return new ZmapCriteria(property, Operator.getPrefixOperator(value), Operator.removePrefixOperator(value),
                        targetType);
            }
        } else if (targetType == String.class) {
            return new ZmapCriteria(property, Operator.getInfixOperator(value), value, targetType);
        }

        return new ZmapCriteria(property, Operator.EQUAL, value, targetType);
    }

    /**
     * Fügt eine Bedingung hinzu (Objekt) Alle Attribute des Objekts sind
     * entweder vom Typ String oder Objekt Für einen String wird ein Kriterium
     * hinzugefügt (s.o.), für ein Objekt wird die Methode rekursiv noch einmal
     * ausgeführt
     * 
     * @param property
     * @param value
     * @param targetTypes
     */
    protected void addCriteria(Path<E> property, Object value, Map<String, Class> targetTypes) {

        // Definition der Praedikate
        Map<String, Object> exmampleProperties = propertyMap(value);
        for (Entry<String, Object> entry : exmampleProperties.entrySet()) {

            String propertyName = entry.getKey();
            Object searchObject = entry.getValue();
            Class targetClass = targetTypes.get(propertyName);
            LOGGER4J.trace("propertyName: " + propertyName);
            LOGGER4J.trace("searchObject: " + searchObject);
            LOGGER4J.trace("targetClass: " + targetClass);

            if (targetClass == null) {
                LOGGER4J.error("Cannot find class for property " + propertyName);
                LOGGER4J.error("Known structure of the target type: ");
                for (String key : targetTypes.keySet()) {
                    LOGGER4J.error("property: " + key + " : class: " + targetTypes.get(key));
                }
                throw new IllegalStateException("Incompatible Example Class");
            }
            if (searchObject == null) {
                continue;
            }

            if (searchObject instanceof String) {
                if (targetTypes.get(propertyName) != null) {
                    addCriteria4String(property.get(propertyName), (String) searchObject, targetClass);
                }
            } else if (searchObject.getClass().isEnum()) {
                addCriteria4Enum(property.get(propertyName), searchObject, targetClass);
            } else {
                addCriteria(property.get(propertyName), searchObject, getTargetTypes(targetClass));
            }
        }
    }

    private ZmapCriteria addCriteria4Enum(Path<Object> property, Object value, Class targetType) {
        return new ZmapCriteria(property, Operator.EQUAL, value, targetType);
    }

    private Map<String, Object> propertyMap(Object o) {
        Map<String, Object> resultValue = new HashMap<>();
        Class clazz = o.getClass();
        Method[] methods = clazz.getDeclaredMethods();
        for (Method method : methods) {
            Object value = null;
            String methodName = method.getName();
            if (methodName.startsWith("get")) {
                // Ermittle Wert der Property
                try {
                    value = method.invoke(o);
                } catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
                    LOGGER4J.log(Level.WARN, e.getMessage(), e);
                }
            }
            if (value == null) {
                continue;
            }

            // Ermittle Property
            String[] properties = methodName.split("get");
            String property = properties[1];
            property = Character.toLowerCase(property.charAt(0)) + (property.length() > 1 ? property.substring(1) : "");
            resultValue.put(property, value);
        }
        return resultValue;
    }

    private static Map<String, Class> getTargetTypes(Class clazz) {
        LOGGER4J.trace("Target types for " + clazz);
        Map<String, Class> propertyTypes = new HashMap<>();
        Method[] methods = clazz.getMethods();
        for (Method method : methods) {
            String methodName = method.getName();
            if (!methodName.startsWith("get")) {
                continue;
            }
            // Ermittle Property
            String[] properties = methodName.split("get");
            String property = properties[1];
            property = Character.toLowerCase(property.charAt(0)) + (property.length() > 1 ? property.substring(1) : "");
            Class returnType = method.getReturnType();
            propertyTypes.put(property, returnType);
            LOGGER4J.trace("    " + property + " " + returnType.getName());
        }
        return propertyTypes;
    }
}
